"""This module provides the implementation of Style-TTS-based text-to-speech pipelines."""

import time
from typing import Any, Dict, List, Literal, Optional

import nltk
import numpy as np
import torch

from senselab.audio.data_structures import Audio
from senselab.utils.data_structures import DeviceType, Language, TorchModel, _select_device_and_dtype
from senselab.utils.data_structures.logging import logger


class StyleTTS2:
    """A class for managing Torch-based StyleTTS2 models."""

    _models: Dict[str, "StyleTTS2"] = {}
    DEF_ALPHA = 0.3
    DEF_BETA = 0.1
    DEF_EMBEDDING_SCALE = 1

    LF_ALPHA = 0.5
    LF_BETA = 0.9
    LF_EMBEDDING_SCALE = 1.5

    ST_ALPHA = 0.5
    ST_BETA = 0.9
    ST_EMBEDDING_SCALE = 1.5
    DIFFUSION_STEPS = 10

    @classmethod
    def _get_style_tts_2_model(
        cls,
        model: Optional[TorchModel] = None,
        language: Optional[Language] = None,
        device: Optional[DeviceType] = None,
        pretrain_data: Optional[Literal["LibriTTS", "LJSpeech"]] = "LibriTTS",
        force_reload: Optional[bool] = False,
    ) -> Any:  # noqa ANN401
        """Get or create a StyleTTS2 model.

        Args:
            model (TorchModel): The Torch model.
                If None, the default model "wilke0818/StyleTTS2-TorchHub:main" is used.
            language (Optional[Language]): The language of the text (default is None).
                The only supported language is "en" for now.
            device (DeviceType): The device to run the model on (default is None). Supported devices are CPU and CUDA.
            pretrain_data (Optional[str]): the dataset that StyleTTS was trained on, currently only supports LibriTTS
                and LJSpeech
            force_reload (Optional[bool]): Whether to require a reload of the Github repository that is being
                used to load the StyleTTS model rather than using a cached version. Defaulting to False for
                efficiency, however, changes to the model and code might not be loaded as a result

        Returns:
            model: The Torch-based StyleTTS2 model.
        """
        if model is None:
            model = TorchModel(path_or_uri="wilke0818/StyleTTS2-TorchHub", revision="main")
        if language == Language(language_code="en"):
            model_name: str = "styletts2"  # This is the default model they have for English.
        else:
            raise NotImplementedError("Only English is supported for now.")
        device, _ = _select_device_and_dtype(
            user_preference=device, compatible_devices=[DeviceType.CUDA, DeviceType.CPU]
        )

        key = f"{model.path_or_uri}-{model.revision}-{language.name}-{device.value}-{pretrain_data}"
        if key not in cls._models or force_reload:
            my_model = torch.hub.load(
                f"{model.path_or_uri}:{model.revision}",
                model_name,
                trust_repo=True,
                force_reload=force_reload,
                pretrain_data=pretrain_data,
            )
            cls._models[key] = my_model
        return cls._models[key]

    @classmethod
    def synthesize_texts_with_style_tts_2(
        cls,
        texts: List[str],
        target_audios: List[Audio],
        target_transcripts: List[Optional[str]],
        model: Optional[TorchModel] = None,
        language: Optional[Language] = None,
        device: Optional[DeviceType] = None,
        pretrain_data: Optional[Literal["LibriTTS", "LJSpeech"]] = "LibriTTS",
        force_reload: Optional[bool] = False,
        alpha: Optional[float] = None,
        beta: Optional[float] = None,
        diffusion_steps: Optional[int] = None,
        embedding_scale: Optional[float] = None,
        t: float = 0.7,
    ) -> List[Audio]:
        """Synthesizes speech from all the provided text samples and target voices+transcripts using StyleTTS2.

        Args:
            texts (List[str]): The list of text strings to be synthesized.
            target_audios (List[Audio]):
                The list of audio objects to reference.
            target_transcripts (List[Optional[str]]):
                Transcript for each target audio
            model (TorchModel): The Torch model.
                If None, the default model "wilke0818/StyleTTS2-TorchHub" is used.
            language (Optional[Language]): The language of the text (default is None).
                The only supported language is "en" for now.
            device (Optional[DeviceType]): device to run model on
            pretrain_data (Optional[Literal]): The dataset (LibriTTS/LJSpeech) the model was trained on
            force_reload (Optional[bool]): whether to reload the repository that the model is coming from
            alpha (float): determines the timbre of the speaker. 0 will match the reference audio
                and 1 will match the content of the transcript. Default will vary based on interpreted TTS type:
                no reference transcripts will use 0.3 (whether or not this is longform text) and .5 otherwise.
            beta (float): beta determines the prosody of the speaker. 0 will match the reference audio
                and 1 will match the content of the transcript. Default will vary based on interpreted TTS type:
                no reference transcript and non-longform will use 0.1, longform text and transcripts will use 0.9.
            diffusion_steps (int): sampler is ancestral, the higher the stpes, the more diverse the samples
                are but the slower the synthesis. Defaults to 10 for all types.
            embedding_scale (float): classifier-free guidance scale. The higher the scale, the more conditional
                the style is to the input text and hence more emotional. Default will vary based on interpreted
                TTS type:
                no reference transcripts will use 1 (whether or not this is longform text) and 1.5 otherwise.
            t (float): how to combine previous and current styles in longform text generation.

        Returns:
            List[Audio]: The list of synthesized audio objects.

        Some tips for best quality:
            - Make sure reference audio is clean and between 2 second and 12 seconds.
            - Play around with alpha, beta, the number of diffusion steps, and the embedding scale for your
                reference audio(s) in order to find a balance between matching the reference voice and having
                the voice sound natural with the text.
            - Default values for these parameters are based off of LibriTTS pre-trained model examples the original
                authors provided.

        The original repo of the model is: https://github.com/yl4579/StyleTTS2.
        """
        if model is None:
            model = TorchModel(path_or_uri="wilke0818/StyleTTS2-TorchHub", revision="main")
        nltk.download("punkt")
        nltk.download("punkt_tab")
        # Take the start time of the model initialization
        start_time_model = time.time()
        my_model = cls._get_style_tts_2_model(model, language, device, pretrain_data, force_reload)

        # Take the end time of the model initialization
        end_time_model = time.time()
        # Print the time taken for initialize the StyleTTS2 model
        elapsed_time_pipeline = end_time_model - start_time_model
        logger.info(f"Time taken to initialize the StyleTTS2 model: {elapsed_time_pipeline:.2f} seconds")

        # Take the start time of text-to-speech synthesis
        start_time_tts = time.time()
        audios = []
        diffusion_steps = diffusion_steps if diffusion_steps else cls.DIFFUSION_STEPS

        for i, text in enumerate(texts):
            sentences = text.split(".")
            target_audio = target_audios[i]

            duration = target_audio.waveform.shape[1] / target_audio.sampling_rate
            if duration < 1 or duration > 12:
                logger.warning(
                    f"Warning: Reference audio at index {i} has a duration of {duration} seconds. "
                    "It is recommended to be between 1 second and 12 seconds."
                )

            if len(sentences) > 2 or len(text) > 200:  # arbitrary but tries to handle false positives and negatives
                # longform text generation
                alpha = alpha if alpha else cls.LF_ALPHA
                beta = beta if beta else cls.LF_BETA
                embedding_scale = embedding_scale if embedding_scale else cls.LF_EMBEDDING_SCALE

                s_ref = my_model._compute_style(target_audio.waveform.squeeze(), target_audio.sampling_rate)
                wavs = []
                s_prev = None
                for sentence in sentences:
                    if sentence.strip() == "":
                        continue
                    sentence += "."  # add it back

                    wav, s_prev = my_model._LFinference(
                        sentence,
                        s_prev,
                        s_ref,
                        alpha=alpha,
                        beta=beta,  # make it more suitable for the text
                        t=t,
                        diffusion_steps=diffusion_steps,
                        embedding_scale=embedding_scale,
                    )
                    wavs.append(wav)
                wav_out = np.concatenate(wavs)

            elif target_transcripts[i]:
                # Considered Style Transfer rather than regular TTS
                alpha = alpha if alpha else cls.ST_ALPHA
                beta = beta if beta else cls.ST_BETA
                embedding_scale = embedding_scale if embedding_scale else cls.ST_EMBEDDING_SCALE

                target_transcript = target_transcripts[i]
                ref_style = my_model._compute_style(target_audio.waveform.squeeze(), target_audio.sampling_rate)
                wav_out = my_model._STinference(
                    text,
                    ref_style,
                    target_transcript,
                    alpha=alpha,
                    beta=beta,
                    diffusion_steps=diffusion_steps,
                    embedding_scale=embedding_scale,
                )
            else:
                alpha = alpha if alpha else cls.DEF_ALPHA
                beta = beta if beta else cls.DEF_BETA
                embedding_scale = embedding_scale if embedding_scale else cls.DEF_EMBEDDING_SCALE

                ref_style = my_model._compute_style(target_audio.waveform.squeeze(), target_audio.sampling_rate)
                wav_out = my_model._inference(
                    text,
                    ref_style,
                    alpha=alpha,
                    beta=beta,
                    diffusion_steps=diffusion_steps,
                    embedding_scale=embedding_scale,
                )
            audios.append(Audio(waveform=wav_out, sampling_rate=24_000))

        # Take the end time of the text-to-speech synthesis
        end_time_tts = time.time()
        # Print the time taken for text-to-speech synthesis
        elapsed_time_tts = end_time_tts - start_time_tts
        logger.info(f"Time taken for synthesizing audios: {elapsed_time_tts:.2f} seconds")

        return audios



"""Data structures relevant for handling audio files and metadata.

The most basic unit is the Audio object which represents the necessary information of a loaded audio
file and its corresponding metadata. Other functionality and abstract data types are provided for
ease of maintaining the codebase and offering consistent public APIs.
"""

import os
import uuid
import warnings
from typing import Dict, Generator, List, Optional, Tuple, Union

import numpy as np
import torch
import torchaudio
from pydantic import BaseModel, Field, ValidationInfo, field_validator

from senselab.utils.constants import SENSELAB_NAMESPACE


class Audio(BaseModel):
    """Pydantic model for audio and its corresponding metadata.

    Pydantic model for audio that holds the necessary attributes, the actual decoded audio data
    and the sampling rate, to work with audio in python. Contains metadata information as needed
    and has a unique identifier for every audio.

    Attributes:
        waveform: The actual audio data read from an audio file, stored as a torch.Tensor
            of shape (num_channels, num_samples)
        sampling_rate: The sampling rate of the audio file
        orig_path_or_id: Optional str for the original path or an ID to track file over time
        metadata: Optional metadata dictionary of information associated with this Audio instance
            (e.g. participant demographics, audio settings, location information)
    """

    waveform: torch.Tensor
    sampling_rate: int
    orig_path_or_id: str | os.PathLike | None = None
    metadata: Dict = Field(default={})
    model_config = {"arbitrary_types_allowed": True}

    @field_validator("waveform", mode="before")
    def convert_to_tensor(
        cls, v: Union[List[float], List[List[float]], np.ndarray, torch.Tensor], _: ValidationInfo
    ) -> torch.Tensor:
        """Converts the audio data to torch.Tensor of shape (num_channels, num_samples)."""
        temporary_tensor = None
        if isinstance(v, list):
            temporary_tensor = torch.tensor(v)
        elif isinstance(v, np.ndarray):
            temporary_tensor = torch.tensor(v)
        elif isinstance(v, torch.Tensor):
            temporary_tensor = v
        else:
            raise ValueError("Unsupported data type")

        if len(temporary_tensor.shape) == 1:
            # make the audio data [channels=1, samples]
            temporary_tensor = temporary_tensor.unsqueeze(0)
        return temporary_tensor.to(torch.float32)

    @classmethod
    def from_filepath(cls, filepath: str | os.PathLike, metadata: Dict = {}) -> "Audio":
        """Creates an Audio instance from an audio file.

        Args:
            filepath: Filepath of the audio file to read from
            metadata: Additional information associated with the audio file
        """
        array, sampling_rate = torchaudio.load(filepath)

        return cls(waveform=array, sampling_rate=sampling_rate, orig_path_or_id=filepath, metadata=metadata)

    def generate_path(self) -> str | os.PathLike:
        """Generate a path like string for this Audio.

        Generates a path like string for the Audio by either utilizing the orig_path_or_id, checking
        if it is a path (has an extension), otherwise using the id if orig_path_or_id is not an ID
        and giving an extension and relative to the current directory.
        """
        if self.orig_path_or_id:
            if os.path.splitext(self.orig_path_or_id)[-1].lower():
                return self.orig_path_or_id
            else:
                return f"{self.orig_path_or_id}"
        else:
            return f"{self.id()}"

    def id(self) -> str:
        """Generate a unique identifier for the Audio.

        Generate a unique identifier for the Audio where equivalent waveforms and sampling
        rates generate the same IDs.

        Returns: String UUID of the Audio generated by an MD5 hash of the waveform and the sampling_rate
        """
        return str(uuid.uuid3(uuid.uuid3(SENSELAB_NAMESPACE, str(self.waveform)), str(self.sampling_rate)))

    def __eq__(self, other: object) -> bool:
        """Overloads the default BaseModel equality to correctly check equivalence, ignoring metadata."""
        if isinstance(other, Audio):
            return self.id() == other.id()
        return False

    def window_generator(self, window_size: int, step_size: int) -> Generator["Audio", None, None]:
        """Creates a sliding window generator for the audio.

        Creates a generator that yields Audio objects corresponding to each window of the waveform
        using a sliding window. The window size and step size are specified in number of samples.
        If the audio waveform doesn't contain an exact number of windows, the remaining samples
        will be included in the last window.

        Args:
            window_size: Size of each window (number of samples).
            step_size: Step size for sliding the window (number of samples).

        Yields:
            Audio: Audio objects corresponding to each window of the waveform.
        """
        if step_size > window_size:
            warnings.warn(
                "Step size is greater than window size. \
                Some of the audio will not be included in the windows."
            )

        num_samples = self.waveform.size(-1)
        current_position = 0

        while current_position < num_samples:
            # Calculate the end position of the window
            end_position = current_position + window_size

            # If the end_position exceeds the number of samples, take the remaining samples
            # This is not necessary since it is done automatically when slicing tensors.
            # However, it is more explicit.
            if end_position > num_samples:
                end_position = num_samples

            # Get the windowed waveform
            window_waveform = self.waveform[:, current_position:end_position]

            # Create a new Audio instance for this window
            window_audio = Audio(
                waveform=window_waveform,
                sampling_rate=self.sampling_rate,
                orig_path_or_id=f"{self.orig_path_or_id}_{current_position}_{end_position}",
                metadata=self.metadata,
            )

            yield window_audio
            current_position += step_size

    def save_to_file(
        self,
        file_path: Union[str, os.PathLike],
        format: Optional[str] = None,
        encoding: Optional[str] = None,
        bits_per_sample: Optional[int] = None,
        buffer_size: int = 4096,
        backend: Optional[str] = None,
        compression: Optional[Union[float, int]] = None,
    ) -> None:
        """Save the `Audio` object to a file using `torchaudio.save`.

        Args:
            file_path (Union[str, os.PathLike]): The path to save the audio file.
            format (Optional[str]): Audio format to use. Valid values include "wav", "ogg", and "flac".
                If None, the format is inferred from the file extension.
            encoding (Optional[str]): Encoding to use. Valid options include "PCM_S", "PCM_U", "PCM_F", "ULAW", "ALAW".
                This is effective for formats like "wav" and "flac".
            bits_per_sample (Optional[int]): Bit depth for the audio file. Valid values are 8, 16, 24, 32, and 64.
            buffer_size (int): Size of the buffer in bytes for processing file-like objects. Default is 4096.
            backend (Optional[str]): I/O backend to use. Valid options include "ffmpeg", "sox", and "soundfile".
                If None, a backend is automatically selected.
            compression (Optional[Union[float, int]]): Compression level for supported formats like "mp3",
            "flac", and "ogg".
                Refer to `torchaudio.save` documentation for specific compression levels.

        Raises:
            ValueError: If the `Audio` waveform is not 2D, or if the sampling rate is invalid.
            RuntimeError: If there is an error saving the audio file.

        Note:
            - https://pytorch.org/audio/master/generated/torchaudio.save.html
        """
        if self.waveform.ndim != 2:
            raise ValueError("Waveform must be a 2D tensor with shape (num_channels, num_samples).")

        if self.sampling_rate <= 0:
            raise ValueError("Sampling rate must be a positive integer.")

        output_dir = os.path.dirname(file_path)
        if not os.access(output_dir, os.W_OK):
            raise RuntimeError(f"Output directory '{output_dir}' is not writable.")

        try:
            if not os.path.exists(output_dir):
                os.makedirs(os.path.dirname(file_path))
            torchaudio.save(
                uri=file_path,
                src=self.waveform,
                sample_rate=self.sampling_rate,
                channels_first=True,
                format=format,
                encoding=encoding,
                bits_per_sample=bits_per_sample,
                buffer_size=buffer_size,
                backend=backend,
                compression=compression,
            )
        except Exception as e:
            raise RuntimeError(f"Error saving audio to file: {e}") from e


def batch_audios(audios: List[Audio]) -> Tuple[torch.Tensor, Union[int, List[int]], List[Dict]]:
    """Batches the Audios together into a single Tensor, keeping individual Audio information separate.

    Batch all of the Audios into a single Tensor of shape (len(audios), num_channels, num_samples).
    Keeps the Audio information related to each sampling rate and metadata separate for each Audio to
    allow for unbatching after running relevant functionality.

    Args:
        audios: List of audios to batch together. NOTE: Should all have the same number of channels
            and is generally advised to have the same sampling rates if running functionality
            that relies on the sampling rate.

    Returns:
        Returns a tuple of a Tensor that will have the shape (len(audios), num_channels, num_samples),
        the sampling rate (an integer if all have the same sampling rate), and a list of each individual
        audio's metadata information.

    Raises:
        RuntimeError: if all of the Audios do not have the same number of channels
    """
    sampling_rates = []
    num_channels_list = []
    lengths = []
    batched_audio = []
    metadatas = []

    for audio in audios:
        sampling_rates.append(audio.sampling_rate)
        num_channels_list.append(audio.waveform.shape[0])  # Assuming waveform shape is (num_channels, num_samples)
        lengths.append(audio.waveform.shape[1])
        metadatas.append(audio.metadata)

    # Check if all audios have the same number of channels
    if len(set(num_channels_list)) != 1:
        raise RuntimeError("All audios must have the same number of channels.")

    # Raise a warning if sampling rates are not the same
    if len(set(sampling_rates)) != 1:
        warnings.warn("Not all sampling rates are the same.", UserWarning)

    # Pad waveforms to the same length
    max_length = max(lengths)
    for audio in audios:
        waveform = audio.waveform
        padding = max_length - waveform.shape[1]
        if padding > 0:
            pad = torch.zeros((waveform.shape[0], padding), dtype=waveform.dtype)
            waveform = torch.cat([waveform, pad], dim=1)
        batched_audio.append(waveform)

    return_sampling_rates: Union[int, List[int]] = (
        int(sampling_rates[0]) if len(set(sampling_rates)) == 1 else sampling_rates
    )

    return torch.stack(batched_audio), return_sampling_rates, metadatas


def unbatch_audios(batched_audio: torch.Tensor, sampling_rates: int | List[int], metadatas: List[Dict]) -> List[Audio]:
    """Unbatches Audios into a List of Audio objects.

    Uses the batched Audios, their respective sampling rates, and their corresponding metadatas to create
    a list of Audios.

    Args:
        batched_audio: torch.Tensor of shape (batch_size, num_channels, num_samples) to unstack
        sampling_rates: The sampling rate of each batched audio if they differ or a single sampling rate for all of them
        metadatas: The respective metadata for each of the batched audios

    Returns:
        List of Audio objects representing each of the Audios that were previously batched together

    Raises:
        ValueError if the batched_audio is not in the correct shape or if the number of batched_audios does not
            match the amount of metadata and sampling rates (if they were provided as a List) that were provided.
    """
    if len(batched_audio.shape) != 3:
        raise ValueError("Expected batched audios to be of shape (batch_size, num_channels, samples)")
    elif batched_audio.shape[0] != len(metadatas) or (
        isinstance(sampling_rates, List) and batched_audio.shape[0] != len(sampling_rates)
    ):
        raise ValueError(
            "Expected sizes of batched_audio, sampling_rates (if provided as a litst) \
                         and metadata to be equal"
        )

    audios = []
    for i in range(len(metadatas)):
        sampling_rate = sampling_rates[i] if isinstance(sampling_rates, List) else sampling_rates
        metadata = metadatas[i]
        audio = batched_audio[i]
        audios.append(Audio(waveform=audio, sampling_rate=sampling_rate, metadata=metadata))
    return audios



""".. include:: ../../README.md"""  # noqa: D415

from multiprocessing import set_start_method

import nest_asyncio

nest_asyncio.apply()

from senselab.utils.data_structures.pydra_helpers import *  # NOQA

try:
    set_start_method("spawn", force=True)
except RuntimeError:
    # Already set
    pass
